/*
 *$Id: rule.cpp,v 1.16 2001/09/14 15:04:27 chooga Exp $
 */
#include "sct_rules/rule_engine_public.h"

#include "rule.h"
#include "validity_checks.h"
#include "logger.h"
#include "syscalls.h"
#include "token.h"

#include <errno.h>
#include <iostream>
#include <list>
#include <strstream>

rule_identifier rule::last_rule_id = 0;

ostream& operator<<(ostream& os, const rule& r)
{
    os<<"rule: "<<r._rule_id<<endl;
    os<<"\tsyscall: "<<r._syscall_id<<endl;
    list<param_conditional*>::const_iterator prit = r._param_cond.begin();
    for(;prit != r._param_cond.end();++prit) os<<**prit;
    list<process_conditional*>::const_iterator pcit = r._proc_cond.begin();
    for(;pcit != r._proc_cond.end();++pcit) os<<**pcit;

    os<<"\taction: ";
    switch (r._act){
    case (LOG):os<<"LOG"<<endl;
	break;
    case (SUSPEND):os<<"SUSPEND"<<endl;
	break;
    case (FAIL):os<<"FAIL"<<endl;
	break;
    case(KILL):os<<"KILL"<<endl;
	break;
    default:
	os<<"unknown action!"<<endl;
	break;
    }
    return os;
}

rule::rule():_rule_id(++last_rule_id),_syscall_id(0),
	     _act(LAST_ACTION), _cond_match(LAST_COND_MATCH){}

rule::rule(const rule& r)
{
    _rule_id = r._rule_id;
    _syscall_id = r._syscall_id;
    _rule_name = r._rule_name;
    _act = r._act;
    //copy the lists of pointers
    list<param_conditional*>::const_iterator prit = r._param_cond.begin();
    for (; prit != r._param_cond.end(); ++prit){
	_param_cond.push_back(new param_conditional(**prit));
    }
     
    list<process_conditional*>::const_iterator pcit = r._proc_cond.begin();
    for (; pcit != r._proc_cond.end(); ++pcit){
	_proc_cond.push_back(new process_conditional(**pcit));
    }
}

rule::~rule()
{
    list<param_conditional*>::iterator prit = _param_cond.begin();
    for (; prit != _param_cond.end(); ++prit){
	delete *prit;
    }
     
    list<process_conditional*>::iterator pcit = _proc_cond.begin();
    for (; pcit != _proc_cond.end(); ++pcit){
	delete *pcit;
    }
}


void rule::set_syscall_identifier(const string& str)
{
    int sid = syscalls::syscall_id(str); 
    if (sid == -1) {
	logger::log(logger::LOG_LEVEL_ERROR,
		    "set_syscall_identifier: '%s' is not a valid syscall id\n",
		    str.c_str());
	throw eUnknownToken();
    }
    _syscall_id = sid;
}

void rule::set_action(const string& str)
{
    if (is_action(str)){
	if (str=="LOG"){
	    _act = LOG;
	}else if (str=="SUSPEND"){
	    _act = SUSPEND;
	}else if (str=="FAIL"){
	    _act = FAIL;
	}else if (str=="KILL"){
	    _act = KILL;
	}
    }else{
	logger::log(logger::LOG_LEVEL_ERROR,
		    "set_action: '%s' is not a valid action!\n",
		    str.c_str());
	throw eUnknownToken();
    }
}

param_conditional* rule::get_param_conditional()
{
    param_conditional* pcond = new param_conditional();
    if (!pcond)
	return NULL;
    _param_cond.push_back(pcond);
    return pcond;
}

process_conditional* rule::get_process_conditional()
{
    process_conditional* pcond = new process_conditional();
    if (!pcond)
	return NULL;
    _proc_cond.push_back(pcond);
    return pcond;
}

void rule::set_condition_matching(const string& str){
    if (!is_condition_matching(str)){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "set_condition_matching: '%s' is not a valid token\n",
		    str.c_str());
	throw eUnknownToken();
	return;
    }
    if (str=="ANY"){
	_cond_match = ANY;
    }else {
	_cond_match = ALL;
    }
}

bool rule::is_valid() const
{
    bool valid = true;
    string rname = (_rule_name.length() ? _rule_name : "(unknown)");
    if (_syscall_id == 0) {
	logger::log(logger::LOG_LEVEL_ERROR,
		    "rule '%s' does not have a valid syscall\n",
		    rname.c_str());
	valid = false;
    }
    if (_rule_name.empty()){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "rule '%s' with id: '%d' has no rule name\n",
		    rname.c_str(), (int)_rule_id);
	valid = false;
    }
    if (_act == LAST_ACTION){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "rule '%s' does not have a valid action\n",
		    rname.c_str());
	valid = false;
    }
    if (!param_conds_valid()){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "rule '%s' has invalid paramaeter condition clause\n",
		    rname.c_str());
	valid = false;
    }
    if (!proc_conds_valid()){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "rule '%s' has invalid process condition clause\n",
		    rname.c_str());
	valid = false;
    }
    if ((_proc_cond.size() != 0) && (_param_cond.size() != 0)){
	if (_cond_match == LAST_COND_MATCH){
	    logger::log(logger::LOG_LEVEL_ERROR,
			"rule '%s' has at least two conditional clauses "
			"and no CONDITIONAL_MATCHING clause\n",
			rname.c_str());
	    valid = false;
	}
    }
    return valid;
}
     


bool rule::proc_conds_valid() const
{
    bool res = true;
    list<process_conditional*>::const_iterator it = _proc_cond.begin();
    for (; it != _proc_cond.end();++it){
	if (!(*it)->is_valid()){
	    strstream buf;
	    buf<<**it;
	    logger::log(logger::LOG_LEVEL_ERROR,
			"found an invalid process conditional clause: ");
	    logger::log(logger::LOG_LEVEL_ERROR,
			"%s\n",
			buf.str());
	    res = false;
	}
    }
    return res;
}

bool rule::param_conds_valid() const
{
    bool res = true;
    list<param_conditional*>::const_iterator it = _param_cond.begin();
    for (; it != _param_cond.end();++it){
	if (!(*it)->is_valid()){
	    strstream buf;
	    buf<<**it;
	    logger::log(logger::LOG_LEVEL_ERROR,
			"found an invalid parameter conditional clause: ");
	    logger::log(logger::LOG_LEVEL_ERROR,
			"%s\n",
			buf.str());
	    res = false;
	}
    }
    return res;
}


static filter_node_t* 
process_field_t_to_var(process_field_t proc);

static filter_node_t*
add_to_filter_chain(filter_node_t* head, filter_node_t* link, 
		    condition_matching_t cond);

static filter_node_t* 
value_type_t_to_filter_node(const value_type_t& val);

static filter_node_t* 
operand_to_filter_node(operand op, filter_node_t* t1, filter_node_t* t2);

// transform this rule into a filter_node 
filter_node_t* rule::filter_node() const
{
    filter_node_t* all = NULL;
    /* build the differet params we want to match against */
    list<param_conditional*>::const_iterator it1 = _param_cond.begin();
    for (; it1 != _param_cond.end(); ++it1)
    {
	/* user specifed location are 1 based, not 0 based */
	unsigned short location = (*it1)->get_location(); 
	assert(location > 0);
	location = (location > 0) ? location - 1 : location;
	
	filter_node_t* t0 = 
	    make_var_indexed_val(VT_PARAM, location, 
				 (*it1)->get_value().type);
	filter_node_t* t1 = 
	    value_type_t_to_filter_node((*it1)->get_value());
	
	filter_node_t* t2 = 
	    operand_to_filter_node((*it1)->get_operand(), t0, t1);
	    
	/* we have the operand and two arguments, now add them to the chain itself */
	all = add_to_filter_chain(all, t2, _cond_match);
    }
   
    /* implement process conditionals matching */
    list<process_conditional*>::const_iterator it2 = _proc_cond.begin();
    for (; it2 != _proc_cond.end(); ++it2)
    {
	filter_node_t* t0 = process_field_t_to_var((*it2)->get_process_field());

	filter_node_t* t1 = 
	    make_long_val((*it2)->get_value());
	filter_node_t* t2 = make_equal(t0, t1);

	all = add_to_filter_chain(all, t2, _cond_match);
    }

    if (logger::get_verbosity_level() <= logger::LOG_LEVEL_INFO)
        print_filter_node(all);

    return all;
}

filter_node_t* 
process_field_t_to_var(process_field_t proc)
{
    /* TODO: more matches! */
    switch(proc){
    case PT_UID: return make_var_val(VT_PROC_UID, DT_LONG); break; /* TODO: get the actual dt size */
    case PT_GID: return make_var_val(VT_PROC_GID, DT_LONG); break;
    case PT_COMM: return make_var_val(VT_PROC_COMM, DT_STRING); break;
    case PT_PID: 
    case PT_PPID:
    case PT_EUID:
    case PT_EGID:
    case PT_LAST_PROCESS_FIELD:
    default:
	logger::log(logger::LOG_LEVEL_ERROR,
		    "bad or unimplemented process conditional "
	            "in process_field_t_to_var: %d\n",
		    (int)proc);
	return NULL;
    }
}

filter_node_t*
add_to_filter_chain(filter_node_t* head, filter_node_t* link, 
		    condition_matching_t cond)
{
    if (!head)
	return link;

    if (!link)
	return head;

    if (cond == ALL)
	return make_logical_and(head, link);

    return make_logical_or(head, link);
}



filter_node_t* 
value_type_t_to_filter_node(const value_type_t& val)
{
    switch (val.type){
    case DT_BOOL:   return make_bool_val(val.val.b); break;
    case DT_CHAR:   return make_char_val(val.val.c); break;
    case DT_INT:    return make_int_val(val.val.i); break;
    case DT_SHORT:  return make_short_val(val.val.s); break;
    case DT_LONG:   return make_long_val(val.val.l); break;
    case DT_FLOAT:  return make_float_val(val.val.f); break;
    case DT_STRING: return make_string_val(val.val.str); break;
    case DT_LAST_TYPE: 
    default:
	logger::log(logger::LOG_LEVEL_ERROR,
		    "bad data type in value_type_t_to_filter_node - %d\n",
		    val.type);
	return NULL;
    }
}

filter_node_t* 
operand_to_filter_node(operand op, filter_node_t* t1, filter_node_t* t2)
{
    switch(op){
    case OT_EQUAL: return make_equal(t1, t2); break;
    case OT_LESS: return make_less_then(t1, t2); break;
    case OT_PATTERN_MATCH: return make_pattern_match(t1, t2); break;
    case OT_GREATER:  return make_greater_then(t1, t2); break; 
    case OT_LESS_EQUAL:
    case OT_GREATER_EQUAL:
    case OT_NOT_EQUAL:
    case OT_PLUS:
    case OT_MINUS:
    case OT_BIT_SHIFT_LEFT:
    case OT_BIT_SHIFT_RIGHT:
    case OT_BIT_AND:
    case OT_BIT_OR:
    case OT_BIT_XOR:
    case OT_BIT_NOT:
    case OT_LOGIC_AND:
    case OT_LOGIC_OR:
    case OT_LOGIC_NOT:
    case OT_LAST_OPERATOR:
    default:
	logger::log(logger::LOG_LEVEL_ERROR,
		    "unimplemented or unknown op '%d' in operand_to_filter_node\n",
		    (int)op);
	return NULL;
    }
}

// get the rule's action
action_t* 
rule::get_action_t() const
{
    switch(_act){
    case LOG:       return make_action_log(4); break;
    case SUSPEND:   return make_action_suspend_process(); break;
    case FAIL:      return make_action_fail_syscall(ENOSYS); break;
    case KILL:      return make_action_kill_process(); break;
    default:
	logger::log(logger::LOG_LEVEL_ERROR,
		    "bad or unimplented action %d\n",
		    (int)_act);
	return NULL;
    }
}
