/*
 *$Id: filter_in_out.c,v 1.1 2001/09/14 14:58:43 chooga Exp $
 */
/*
 * Rule engine - filter input/output and serialization routines.
 */

#include "mem_manager_public.h"
#include "encode_decode_public.h"
#include "rule_engine_private.h"

/* number of spaces for each indent level. */
#define	INDENT_LEVEL	1


/* operators symbols, used for printing. */
const char* operators_symbols[] = {
	"=", "<", ">", "<=", ">=", "!=", "+", "-", "<<", ">>",
	"&", "|", "^", "!", "&&", "||", "!", "~=",
};

/*
 * Forward declarations.
 */
void do_print_filter_node(struct tracker_filter_node* f, int indent);

/*
 * Private functions.
 */

/* print spaces in the ammount given. */
void
print_indent(int len)
{
	char buf[200];
	char* p_c = buf;

	for ( ; len && p_c < buf+199; len--) {
		*p_c++ = ' ';
	}
	*p_c = '\0';
	PPRINT(buf);
}

/* print a var node. */
void
print_filter_node_var(struct tracker_filter_node* f, int indent)
{
	print_indent(indent);
	PPRINT("var '%d'", f->u.var.type);
	PPRINT(" index '%d'", f->u.var.var_index);
	PPRINT("\n");
}

/* print a const node. */
void
print_filter_node_const(struct tracker_filter_node* f, int indent)
{
	print_indent(indent);
	switch (CONST_DATA_TYPE(f)) {
		case DT_BOOL:
			PPRINT("bool: %d\n", CONST_BOOL(f));
			break;
		case DT_CHAR:
			PPRINT("char: %c (%d)\n",
			       CONST_CHAR(f), (int)CONST_CHAR(f));
			break;
		case DT_INT:
			PPRINT("int: %d\n", CONST_INT(f));
			break;
		case DT_SHORT:
			PPRINT("short: %d\n", CONST_SHORT(f));
			break;
		case DT_LONG:
			PPRINT("long: %ld\n", CONST_LONG(f));
			break;
		case DT_FLOAT:
			PPRINT("float: %f\n", CONST_FLOAT(f));
			break;
		case DT_STRING:
			{
				const char* str = CONST_STRING(f);
				if (str)
					PPRINT("str: %s\n", str);
				else
					PPRINT("str: (nil)\n");
			}
			break;
		default:
			/* shouldn't happen. */
			PPRINT("(unknown type)\n");
			break;
	}
}

/* print a compound node. */
void
print_filter_node_compound(struct tracker_filter_node* f, int indent)
{
	const char* op = "(error)";

	if (f->u.compound.op >= 0 && f->u.compound.op < OT_LAST_OPERATOR)
		op = operators_symbols[f->u.compound.op];
	print_indent(indent);
	PPRINT("operator: '%s' (%d)\n", op, f->u.compound.op);
	print_indent(indent);
	PPRINT("f1:\n");
	if (f->u.compound.node1) {
		do_print_filter_node(f->u.compound.node1, indent+INDENT_LEVEL);
	}
	else {
		print_indent(indent);
		PPRINT("Empty\n");
	}
	print_indent(indent);
	PPRINT("f2:\n");
	if (f->u.compound.node2) {
		do_print_filter_node(f->u.compound.node2, indent+INDENT_LEVEL);
	}
	else {
		print_indent(indent);
		PPRINT("Empty\n");
	}
}

/* recursively print the given filter node, beginning */
/* with the given indentation level.                  */
void
do_print_filter_node(struct tracker_filter_node* f, int indent)
{
	switch(f->type) {
		case NT_VAR:
			print_filter_node_var(f, indent);
			break;
		case NT_CONST:
			print_filter_node_const(f, indent);
			break;
		case NT_COMPOUND:
			print_filter_node_compound(f, indent);
			break;
		default:
			/* shouldn't happen. */
			print_indent(indent);
			PPRINT("Error: invalid node type '%d'\n", f->type);
			break;
	}
}


/* print the given filter node (and all its sub-nodes) to output. */
/* output is stdout if in user-mode, or using printk if in kernel */
/* mode. the output is being formatted to reflext the hierarchy   */
/* of the expression stored in the tree - good for debugging.     */
void
print_filter_node(filter_node_t* f)
{
	int indent = 3;
	if (f) {
		PPRINT("  Filter node data:\n");
		do_print_filter_node(f, indent);
		PPRINT("\n");
	}
	else {
		PPRINT("Empty filter node\n");
	}
}


/*
 * Filter serialization/de-serialization functions.
 */


/*
 * Private functions.
 */

/* serialize a 'var' filter node. */
int
serialize_filter_node_var(filter_node_t* f,
			  unsigned char* buf,
			  int buflen,
			  unsigned char** pos)
{
	int rc = encode_char(buf, buflen, pos, NT_VAR);
	if (rc > 0) {
		rc = encode_char(buf, buflen, pos, f->data_type);
	}
	if (rc > 0) {
		rc = encode_char(buf, buflen, pos, f->u.var.type);
	}
	if (rc > 0) {
		rc = encode_int(buf, buflen, pos, f->u.var.var_index);
	}

	return rc;
}

/* de-serialize a 'var' filter node. */
int
deserialize_filter_node_var(filter_node_t** pf,
			    unsigned char* buf,
			    int buflen,
			    unsigned char** pos)
{
	char var_data_type;
	char var_type;
	int var_index;
	int rc = decode_char(buf, buflen, pos, &var_data_type);

	if (rc > 0) {
		rc = decode_char(buf, buflen, pos, &var_type);
	}
	if (rc > 0) {
		rc = decode_int(buf, buflen, pos, &var_index);
	}
	if (rc > 0) {
		*pf = make_var_indexed_val(var_type, var_index, var_data_type);
		if (!*pf) {
			rc = 0;
		}
	}

	return rc;
}

/* serialize a 'const' filter node. */
int
serialize_filter_node_const(filter_node_t* f,
			    unsigned char* buf,
			    int buflen,
			    unsigned char** pos)
{
	int rc = encode_char(buf, buflen, pos, NT_CONST);
	if (rc > 0) {
		rc = encode_char(buf, buflen, pos, f->data_type);
	}
	if (rc > 0) {
		rc = encode_char(buf, buflen, pos, f->u.constant.type);
	}
	if (rc > 0) {
		switch (f->u.constant.type) {
			case DT_BOOL:
				rc = encode_char(buf,
						 buflen, pos, CONST_BOOL(f));
				break;
			case DT_CHAR:
				rc = encode_char(buf,
						 buflen, pos, CONST_CHAR(f));
				break;
			case DT_SHORT:
				rc = encode_short(buf,
						  buflen, pos, CONST_SHORT(f));
				break;
			case DT_INT:
				rc = encode_int(buf,
						buflen, pos, CONST_INT(f));
				break;
			case DT_LONG:
				rc = encode_long(buf,
						 buflen, pos, CONST_LONG(f));
				break;
			case DT_FLOAT:
				rc = encode_float(buf,
						  buflen, pos, CONST_FLOAT(f));
				break;
			case DT_STRING:
				rc = encode_string(buf,
						  buflen, pos, CONST_STRING(f));
				break;
			default:
				/* shouldn't happen. */
				rc = 0;
				break;
		}
	}

	return rc;
}

/* de-serialize a 'const' filter node. */
int
deserialize_filter_node_const(filter_node_t** pf,
			      unsigned char* buf,
			      int buflen,
			      unsigned char** pos)
{
	char data_type;
	char constant_type;
	int rc = decode_char(buf, buflen, pos, &data_type);

	if (rc > 0) {
		rc = decode_char(buf, buflen, pos, &constant_type);
	}
	if (rc > 0) {
		switch (constant_type) {
			case DT_BOOL:
				{
					char c;
					rc = decode_char(buf,
						 	 buflen, pos, &c);
					if (rc > 0) {
						*pf = make_bool_val(c);
						if (!*pf) {
							rc = 0;
						}
					}
				}
				break;
			case DT_CHAR:
				{
					char c;
					rc = decode_char(buf,
						 	 buflen, pos, &c);
					if (rc > 0) {
						*pf = make_char_val(c);
						if (!*pf) {
							rc = 0;
						}
					}
				}
				break;
			case DT_SHORT:
				{
					short s;
					rc = decode_short(buf,
						 	  buflen, pos, &s);
					if (rc > 0) {
						*pf = make_short_val(s);
						if (!*pf) {
							rc = 0;
						}
					}
				}
				break;
			case DT_INT:
				{
					int i;
					rc = decode_int(buf,
						 	buflen, pos, &i);
					if (rc > 0) {
						*pf = make_int_val(i);
						if (!*pf) {
							rc = 0;
						}
					}
				}
				break;
			case DT_LONG:
				{
					long l;
					rc = decode_long(buf,
						 	 buflen, pos, &l);
					if (rc > 0) {
						*pf = make_long_val(l);
						if (!*pf) {
							rc = 0;
						}
					}
				}
				break;
			case DT_FLOAT:
				{
					float f;
					rc = decode_float(buf,
						 	  buflen, pos, &f);
					if (rc > 0) {
						*pf = make_float_val(f);
						if (!*pf) {
							rc = 0;
						}
					}
				}
				break;
			case DT_STRING:
				{
					char* str;
					rc = decode_string(buf,
						 	   buflen, pos, &str);
					if (rc > 0) {
						*pf = make_string_val(str);
						if (!*pf) {
							rc = 0;
						}
						mem_free(str);
					}
				}
				break;
			default:
				/* shouldn't happen. */
				rc = 0;
				break;
		}
	}

	return rc;
}

/* serialize a 'compound' filter node, and its sub-nodes. */
int
serialize_filter_node_compound(filter_node_t* f,
			       unsigned char* buf,
			       int buflen,
			       unsigned char** pos)
{
	int rc = encode_char(buf, buflen, pos, NT_COMPOUND);
	if (rc > 0) {
		rc = encode_char(buf, buflen, pos, f->data_type);
	}
	if (rc > 0) {
		rc = encode_char(buf, buflen, pos, f->u.compound.op);
	}
	if (rc > 0) {
		rc = do_serialize_filter_node(f->u.compound.node1,
					      buf, buflen, pos);
	}
	if (rc > 0) {
		rc = do_serialize_filter_node(f->u.compound.node2,
					      buf, buflen, pos);
	}

	return rc;
}

/* de-serialize a 'compound' filter node, and its sub-nodes. */
int
deserialize_filter_node_compound(filter_node_t** pf,
			         unsigned char* buf,
			         int buflen,
			         unsigned char** pos)
{
	char data_type;
	char op;
	filter_node_t* f1 = NULL;
	filter_node_t* f2 = NULL;
	int rc = decode_char(buf, buflen, pos, &data_type);

	if (rc > 0) {
		rc = decode_char(buf, buflen, pos, &op);
	}
	if (rc > 0) {
		rc = do_deserialize_filter_node(&f1, buf, buflen, pos);
	}
	if (rc > 0) {
		rc = do_deserialize_filter_node(&f2, buf, buflen, pos);
	}
	if (rc > 0) {
		*pf = make_compound_node_with_type(op, f1, f2, data_type);
		if (!*pf) {
			rc = 0;
		}
	}

	/* free the sub-nodes, if we failed somewhere along the path. */
	if (rc < 1) {
		if (f1)
			free_filter_node(f1);
		if (f2)
			free_filter_node(f2);
	}

	return rc;
}

/* serialize a filter node (and all its sub-nodes) into a */
/* buffer, for export purposes. returns the size of the   */
/* serialized data, 0 if there was an error, -1 if buffer */
/* is too small. serialization begins at 'pos' in 'buf'.  */
int
do_serialize_filter_node(filter_node_t* f,
		         unsigned char* buf,
		         int buflen,
		         unsigned char** pos)
{
	int rc = 0;

	if (f) {
		switch (f->type) {
			case NT_VAR:
				rc = serialize_filter_node_var(f, buf, buflen, pos);
				break;
			case NT_CONST:
				rc = serialize_filter_node_const(f, buf, buflen, pos);
				break;
			case NT_COMPOUND:
				rc = serialize_filter_node_compound(f, buf, buflen, pos);
				break;
			default:
				/* shouldn't happen. */
				rc = 0;
				break;
		}
	}
	else {
		rc = encode_char(buf, buflen, pos, NT_NIL);
	}

	return rc;
}

/* de-serialize a filter node (and all its sub-nodes)     */
/* from a buffer, for import purposes. returns the new    */
/* allocated filter node (the top of the tree), or NULL   */
/* if an error occured. de-serialization begins as 'pos'  */
/* in 'buf'.                                              */
int
do_deserialize_filter_node(filter_node_t** pf,
			   unsigned char* buf,
			   int buflen,
			   unsigned char** pos)
{
	int rc = 0;
	char nt;

	rc = decode_char(buf, buflen, pos, &nt);
	if (rc > 0) {
		switch (nt) {
			case NT_VAR:
				rc = deserialize_filter_node_var(pf, buf, buflen, pos);
				break;
			case NT_CONST:
				rc = deserialize_filter_node_const(pf, buf, buflen, pos);
				break;
			case NT_COMPOUND:
				rc = deserialize_filter_node_compound(pf, buf, buflen, pos);
				break;
			case NT_NIL:
				/* a NULL filter_node. */
				*pf = NULL;
				rc = 1;
				break;
			default:
				/* shouldn't happen. */
				rc = 0;
				break;
		}
	}

	return rc;
}

/*
 * Public interface implementation.
 */

/* serialize a filter node (and all its sub-nodes) into a */
/* buffer, for export purposes. returns the size of the   */
/* serialized data, 0 if there was an error, -1 if buffer */
/* is too small.                                          */
int
serialize_filter_node(filter_node_t* f, unsigned char* buf, int buflen)
{
	unsigned char* pos = buf;
	int rc = do_serialize_filter_node(f, buf, buflen, &pos);

	return (rc > 0 ? pos - buf : rc);
}

/* de-serialize a filter node (and all its sub-nodes)     */
/* from a buffer, for import purposes. returns the new    */
/* allocated filter node (the top of the tree), or NULL   */
/* if an error occured.                                   */
filter_node_t*
deserialize_filter_node(unsigned char* buf, int buflen)
{
	unsigned char* pos = buf;
	filter_node_t* f = NULL;
	int rc = 0;

	if (buflen > 0) {
		rc = do_deserialize_filter_node(&f, buf, buflen, &pos);
	}

	return (rc > 0 ? f : NULL);
}

