/*
 *$Id: filter_node_build.c,v 1.1 2001/09/14 14:58:43 chooga Exp $
 */
/*
 * Rule engine - implementation of the filtering nodes building.
 */

#include "rule_engine_private.h"
#include "mem_manager_public.h"

#include "archdep.h"
/*
 * Private functions.
 */


/* build a compound node. */
struct tracker_filter_node*
make_compound_node_with_type(oper_type_t op,
		   	     struct tracker_filter_node* node1,
		   	     struct tracker_filter_node* node2,
		   	     data_type_t data_type)
{
	struct tracker_filter_node* node =
		(struct tracker_filter_node*)
			mem_alloc(sizeof(struct tracker_filter_node));
	if (node) {
		node->type = NT_COMPOUND;
		node->data_type = data_type;
		node->u.compound.op = op;
		node->u.compound.node1 = node1;
		node->u.compound.node2 = node2;
	}

	return node;
}

/* build a compound node. deduce the data type from its first node param. */
static struct tracker_filter_node*
make_compound_node(oper_type_t op,
		   struct tracker_filter_node* node1,
		   struct tracker_filter_node* node2)
{
	return make_compound_node_with_type(op, node1, node2, node1->data_type);
}

/* free the 'var' part of a filter node. don't free the node itself. */
static void
free_var_of_filter_node(struct tracker_filter_node* f)
{
	/* nothing to do so far - var part contains only numbers. */
}

/* free the 'const' part of a filter node. don't free the node itself. */
static void
free_const_of_filter_node(struct tracker_filter_node* f)
{
	/* free string, in case of a string type node. */
	if (f->u.constant.type == DT_STRING) {
		if (f->u.constant.u.str) {
			mem_free(f->u.constant.u.str);
			f->u.constant.u.str = NULL;
		}
	}
}

/* free the 'compound' part of a filter node. don't free the node itself. */
static void
free_compound_of_filter_node(struct tracker_filter_node* f)
{
	if (f->u.compound.node1) {
		free_filter_node(f->u.compound.node1);
		f->u.compound.node1 = NULL;
	}
	if (f->u.compound.node2) {
		free_filter_node(f->u.compound.node2);
		f->u.compound.node2 = NULL;
	}
}


/*
 * Implementation of public filter building functions.
 */

/*
 * functions that are used to build a filtering expression.
 */

/* create primitive types - int, string, variable, indexed variable */
/* (e.g. parameter).                                                */
filter_node_t*
make_bool_val(int num)
{
	struct tracker_filter_node* node =
		(struct tracker_filter_node*)
			mem_alloc(sizeof(struct tracker_filter_node));
	if (node) {
		node->type = NT_CONST;
		node->data_type = DT_BOOL;
		node->u.constant.type = DT_BOOL;
		node->u.constant.u.b = (num ? TF_TRUE : TF_FALSE);
	}

	return node;
}

filter_node_t*
make_int_val(int num)
{
	struct tracker_filter_node* node =
		(struct tracker_filter_node*)
			mem_alloc(sizeof(struct tracker_filter_node));
	if (node) {
		node->type = NT_CONST;
		node->data_type = DT_INT;
		node->u.constant.type = DT_INT;
		node->u.constant.u.i = num;
	}

	return node;
}

filter_node_t*
make_char_val(char c)
{
	struct tracker_filter_node* node =
		(struct tracker_filter_node*)
			mem_alloc(sizeof(struct tracker_filter_node));
	if (node) {
		node->type = NT_CONST;
		node->data_type = DT_CHAR;
		node->u.constant.type = DT_CHAR;
		node->u.constant.u.c = c;
	}

	return node;
}

filter_node_t*
make_short_val(short num)
{
	struct tracker_filter_node* node =
		(struct tracker_filter_node*)
			mem_alloc(sizeof(struct tracker_filter_node));
	if (node) {
		node->type = NT_CONST;
		node->data_type = DT_SHORT;
		node->u.constant.type = DT_SHORT;
		node->u.constant.u.s = num;
	}

	return node;
}

filter_node_t*
make_long_val(long num)
{
	struct tracker_filter_node* node =
		(struct tracker_filter_node*)
			mem_alloc(sizeof(struct tracker_filter_node));
	if (node) {
		node->type = NT_CONST;
		node->data_type = DT_LONG;
		node->u.constant.type = DT_LONG;
		node->u.constant.u.l = num;
	}

	return node;
}

filter_node_t*
make_float_val(float num)
{
	struct tracker_filter_node* node =
		(struct tracker_filter_node*)
			mem_alloc(sizeof(struct tracker_filter_node));
	if (node) {
		node->type = NT_CONST;
		node->data_type = DT_FLOAT;
		node->u.constant.type = DT_FLOAT;
		node->u.constant.u.f = num;
	}

	return node;
}

filter_node_t*
make_string_val(const char* str)
{
	struct tracker_filter_node* node =
		(struct tracker_filter_node*)
			mem_alloc(sizeof(struct tracker_filter_node));
	if (node) {
		node->type = NT_CONST;
		node->data_type = DT_STRING;
		node->u.constant.type = DT_STRING;
		if (str) {
			node->u.constant.u.str =
				(char*)mem_alloc(strlen(str)+1);
			if (node->u.constant.u.str)
				strcpy(node->u.constant.u.str, str);
			else {
				mem_free(node);
				node = NULL;
			}
		}
		else {
			node->u.constant.u.str = NULL;
		}
	}

	return node;
}

filter_node_t*
make_var_val(var_type_t v, data_type_t var_data_type)
{
	struct tracker_filter_node* node =
		(struct tracker_filter_node*)
			mem_alloc(sizeof(struct tracker_filter_node));
	if (node) {
		node->type = NT_VAR;
		node->data_type = var_data_type;
		node->u.var.type = v;
		node->u.var.var_index = 0;
	}

	return node;
}

filter_node_t*
make_var_indexed_val(var_type_t v, int var_index, data_type_t var_data_type)
{
	struct tracker_filter_node* node =
		(struct tracker_filter_node*)
			mem_alloc(sizeof(struct tracker_filter_node));
	if (node) {
		node->type = NT_VAR;
		node->data_type = var_data_type;
		node->u.var.type = v;
		node->u.var.var_index = var_index;
	}

	return node;
}


/* combine expressions using operators.                          */
/* checks for matching (e.g. can't comprate and int and an str). */
filter_node_t*
make_equal(filter_node_t* f1, filter_node_t* f2)
{
	filter_node_t* f = NULL;

	if (f1 && f2) {
		f = make_compound_node_with_type(OT_EQUAL, f1, f2, DT_BOOL);
		if (!f) {
			free_filter_node(f1);
			free_filter_node(f2);
		}
	}
	else {
		free_filter_node(f1);
		free_filter_node(f2);
	}

	return f;
}

filter_node_t*
make_less_then(filter_node_t* f1, filter_node_t* f2)
{
	filter_node_t* f = NULL;

	if (f1 && f2)
		f = make_compound_node_with_type(OT_LESS, f1, f2, DT_BOOL);
		if (!f) {
			free_filter_node(f1);
			free_filter_node(f2);
		}
	else {
		free_filter_node(f1);
		free_filter_node(f2);
	}

	return f;
}

filter_node_t*
make_less_or_equal(filter_node_t* f1, filter_node_t* f2)
{
	filter_node_t* f = NULL;

	if (f1 && f2) {
		f = make_compound_node_with_type(OT_LESS_EQUAL, f1, f2, DT_BOOL);
		if (!f) {
			free_filter_node(f1);
			free_filter_node(f2);
		}
	}
	else {
		free_filter_node(f1);
		free_filter_node(f2);
	}

	return f;
}

filter_node_t*
make_greater_then(filter_node_t* f1, filter_node_t* f2)
{
	filter_node_t* f = NULL;

	if (f1 && f2) {
		f = make_compound_node_with_type(OT_GREATER, f1, f2, DT_BOOL);
		if (!f) {
			free_filter_node(f1);
			free_filter_node(f2);
		}
	}
	else {
		free_filter_node(f1);
		free_filter_node(f2);
	}

	return f;
}

filter_node_t*
make_greater_or_equal(filter_node_t* f1, filter_node_t* f2)
{
	filter_node_t* f = NULL;

	if (f1 && f2) {
		f = make_compound_node_with_type(OT_GREATER_EQUAL,
						 f1, f2, DT_BOOL);
		if (!f) {
			free_filter_node(f1);
			free_filter_node(f2);
		}
	}
	else {
		free_filter_node(f1);
		free_filter_node(f2);
	}

	return f;
}

filter_node_t*
make_not_equal(filter_node_t* f1, filter_node_t* f2)
{
	filter_node_t* f = NULL;

	if (f1 && f2) {
		f = make_compound_node_with_type(OT_NOT_EQUAL, f1, f2, DT_BOOL);
		if (!f) {
			free_filter_node(f1);
			free_filter_node(f2);
		}
	}
	else {
		free_filter_node(f1);
		free_filter_node(f2);
	}

	return f;
}

filter_node_t*
make_logical_and(filter_node_t* f1, filter_node_t* f2)
{
	filter_node_t* f = NULL;

	if (f1 && f2) {
		f = make_compound_node_with_type(OT_LOGIC_AND, f1, f2, DT_BOOL);
		if (!f) {
			free_filter_node(f1);
			free_filter_node(f2);
		}
	}
	else {
		free_filter_node(f1);
		free_filter_node(f2);
	}

	return f;
}

filter_node_t*
make_logical_or(filter_node_t* f1, filter_node_t* f2)
{
	filter_node_t* f = NULL;

	if (f1 && f2) {
		f = make_compound_node_with_type(OT_LOGIC_OR, f1, f2, DT_BOOL);
		if (!f) {
			free_filter_node(f1);
			free_filter_node(f2);
		}
	}
	else {
		free_filter_node(f1);
		free_filter_node(f2);
	}

	return f;
}

filter_node_t*
make_logical_not(filter_node_t* f1)
{
	filter_node_t* f = NULL;

	if (f1) {
		f = make_compound_node_with_type(OT_LOGIC_NOT,
						 f1, NULL, DT_BOOL);
		if (!f) {
			free_filter_node(f1);
		}
	}
	else {
		free_filter_node(f1);
	}

	return f;
}

filter_node_t*
make_bit_and(filter_node_t* f1, filter_node_t* f2)
{
	filter_node_t* f = NULL;

	if (f1 && f2 &&
	    f1->data_type != DT_STRING && f2->data_type != DT_STRING) {
		f = make_compound_node(OT_BIT_AND, f1, f2);
		if (!f) {
			free_filter_node(f1);
			free_filter_node(f2);
		}
	}
	else {
		free_filter_node(f1);
		free_filter_node(f2);
	}

	return f;
}

filter_node_t*
make_bit_or(filter_node_t* f1, filter_node_t* f2)
{
	filter_node_t* f = NULL;

	if (f1 && f2 &&
	    f1->data_type != DT_STRING && f2->data_type != DT_STRING) {
		f = make_compound_node(OT_BIT_OR, f1, f2);
		if (!f) {
			free_filter_node(f1);
			free_filter_node(f2);
		}
	}
	else {
		free_filter_node(f1);
		free_filter_node(f2);
	}

	return f;
}

filter_node_t*
make_bit_xor(filter_node_t* f1, filter_node_t* f2)
{
	filter_node_t* f = NULL;

	if (f1 && f2 &&
	    f1->data_type != DT_STRING && f2->data_type != DT_STRING) {
		f = make_compound_node(OT_BIT_XOR, f1, f2);
		if (!f) {
			free_filter_node(f1);
			free_filter_node(f2);
		}
	}
	else {
		free_filter_node(f1);
		free_filter_node(f2);
	}

	return f;
}

filter_node_t*
make_bit_not(filter_node_t* f1)
{
	filter_node_t* f = NULL;

	if (f1 && f1->data_type != DT_STRING) {
		f = make_compound_node(OT_BIT_NOT, f1, NULL);
		if (!f) {
			free_filter_node(f1);
		}
	}
	else {
		free_filter_node(f1);
	}

	return f;
}

filter_node_t*
make_bit_left_shift(filter_node_t* f1, filter_node_t* f2)
{
	filter_node_t* f = NULL;

	if (f1 && f2 &&
	    f1->data_type != DT_STRING && f2->data_type != DT_STRING) {
		f = make_compound_node(OT_BIT_SHIFT_LEFT, f1, f2);
		if (!f) {
			free_filter_node(f1);
			free_filter_node(f2);
		}
	}
	else {
		free_filter_node(f1);
		free_filter_node(f2);
	}

	return f;
}

filter_node_t*
make_bit_right_shift(filter_node_t* f1, filter_node_t* f2)
{
	filter_node_t* f = NULL;

	if (f1 && f2 &&
	    f1->data_type != DT_STRING && f2->data_type != DT_STRING) {
		f = make_compound_node(OT_BIT_SHIFT_RIGHT, f1, f2);
		if (!f) {
			free_filter_node(f1);
			free_filter_node(f2);
		}
	}
	else {
		free_filter_node(f1);
		free_filter_node(f2);
	}

	return f;
}

filter_node_t*
make_math_plus(filter_node_t* f1, filter_node_t* f2)
{
	filter_node_t* f = NULL;

	if (f1 && f2 &&
	    f1->data_type != DT_STRING && f2->data_type != DT_STRING) {
		f = make_compound_node(OT_PLUS, f1, f2);
		if (!f) {
			free_filter_node(f1);
			free_filter_node(f2);
		}
	}
	else {
		free_filter_node(f1);
		free_filter_node(f2);
	}

	return f;
}

filter_node_t*
make_math_minus(filter_node_t* f1, filter_node_t* f2)
{
	filter_node_t* f = NULL;

	if (f1 && f2 &&
	    f1->data_type != DT_STRING && f2->data_type != DT_STRING) {
		f = make_compound_node(OT_MINUS, f1, f2);
		if (!f) {
			free_filter_node(f1);
			free_filter_node(f2);
		}
	}
	else {
		free_filter_node(f1);
		free_filter_node(f2);
	}

	return f;
}

filter_node_t*
make_pattern_match(filter_node_t* f1, filter_node_t* f2)
{
	filter_node_t* f = NULL;

	if (f1 && f2 &&
	    f1->data_type == DT_STRING && f2->data_type == DT_STRING) {
		f = make_compound_node(OT_PATTERN_MATCH, f1, f2);
		if (!f) {
			free_filter_node(f1);
			free_filter_node(f2);
		}
	}
	else {
		free_filter_node(f1);
		free_filter_node(f2);
	}

	return f;
}


/* free the given filter node, as well as all sub-nodes and */
/* heap memory it contains, directly and indirectly.        */
void
free_filter_node(filter_node_t* f)
{
	if (f) {
		/* free contents, based on node type. */
		switch (f->type) {
			case NT_VAR:
				free_var_of_filter_node(f);
				break;
			case NT_CONST:
				free_const_of_filter_node(f);
				break;
			case NT_COMPOUND:
				free_compound_of_filter_node(f);
				break;
			default:
				/* shouldn't happen. */
				break;
		}
		mem_free(f);
	}
}

