/*
 *$Id: rule_engine.c,v 1.1 2001/09/14 14:58:43 chooga Exp $
 */
#include "mem_manager_public.h"
#include "encode_decode_public.h"
#include "rule_engine_private.h"
#include "filter_vars_public.h"
#include "archdep.h"

#include "stdarg.h"

/* a list of tracking rules. */
#define	MAX_NUM_TRACKER_RULES	20

struct rule_list {
    struct tracker_rule* rules[MAX_NUM_TRACKER_RULES];
    int num_rules;
};


/*
 * Private functions.
 */

/* find the location of a rule with the given ID in the rules list. */
/* returns the rule's location, or '-1' if none is found.           */
int
find_tracking_rule(rule_list_t* rules, int rule_id)
{
	int i;

	if (!rules)
	    return -1;

	for (i = 0; i < rules->num_rules; i++) {
		if (rules->rules[i] && rules->rules[i]->rule_id == rule_id)
			break;
	}

	return (i < rules->num_rules ? i : -1);
}

/* add the given rule to the rules list.                                */
/* returns the rule's location on success, or '-1' if an error occured. */
int
do_add_tracking_rule(rule_list_t* rules,
		     int rule_id,
		     filter_node_t* filter,
		     action_t* action)
{
	int i;
	int j;
	struct tracker_rule* rule;

	/* make sure we have free space. */
	if (!rules || rules->num_rules >= MAX_NUM_TRACKER_RULES)
		return -1;

	/* allocate a structure for the new rule. */
	rule = mem_alloc(sizeof(struct tracker_rule));
	if (!rule)
	    return -1;

	rule->rule_id = rule_id;
	rule->filter = filter;
	/* TODO - write a function to make proper copying of the action. */
	memcpy(&rule->action, action, sizeof(struct tracker_action));
	
	/* find the right location for the new rule. */
	for (i = 0; i < rules->num_rules; i++) {
	    if (rules->rules[i] && rules->rules[i]->rule_id > rule_id)
		break;
	}
	
	/* make a free space for the new rule. */
	for (j = rules->num_rules; j > i; j--) {
	    rules->rules[j] = rules->rules[j-1];
	}
	
	/* insert the new rule. */
	rules->rules[i] = rule;
	rules->num_rules++;
	
	return (i);
}

/* serialize a 'action_t' action node. */
int
do_serialize_action_node(action_t* a,
			 unsigned char* buf,
			 int buflen,
			 unsigned char** pos)
{
	int rc = 0;

	if (a) {
		char action_type = a->type;
		rc = encode_char(buf, buflen, pos, action_type);
		if (rc > 0) {
			switch (action_type) {
				case AT_LOG:
					rc = encode_int(buf,
							buflen, pos, a->u.priority);
					break;
				case AT_RULE_ENABLE:
					rc = encode_int(buf,
							buflen, pos, a->u.rule_id);
					break;
				case AT_RULE_DISABLE:
					rc = encode_int(buf,
							buflen, pos, a->u.rule_id);
					break;
				case AT_PROCESS_SUSPEND:
					/* no data for this action. */
					break;
				case AT_PROCESS_KILL:
					/* no data for this action. */
					break;
				case AT_SYSCALL_FAIL:
					rc = encode_int(buf,
							buflen, pos, a->u.error_code);
					break;
				default:
					rc = 0;
					break;
			}
		}
	}
	else {
		rc = encode_char(buf, buflen, pos, AT_NIL);
	}

	return rc;
}

/* de-serialize an 'action_t' action node. */
int
do_deserialize_action_node(action_t** pa,
			   unsigned char* buf,
			   int buflen,
			   unsigned char** pos)
{
	char action_type;
	int rc = decode_char(buf, buflen, pos, &action_type);

	if (rc > 0) {
		switch (action_type) {
			case AT_LOG:
				{
					int priority;
					rc = decode_int(buf,
						 	buflen, pos, &priority);
					if (rc > 0) {
						*pa = make_action_log(priority);
						if (!*pa) {
							rc = 0;
						}
					}
				}
				break;
			case AT_RULE_ENABLE:
				{
					int rule_id;
					rc = decode_int(buf,
						 	buflen, pos, &rule_id);
					if (rc > 0) {
						*pa = make_action_enable_rule(
								       rule_id);
						if (!*pa) {
							rc = 0;
						}
					}
				}
				break;
			case AT_RULE_DISABLE:
				{
					int rule_id;
					rc = decode_int(buf,
						 	buflen, pos, &rule_id);
					if (rc > 0) {
						*pa = make_action_disable_rule(
								       rule_id);
						if (!*pa) {
							rc = 0;
						}
					}
				}
				break;
			case AT_PROCESS_SUSPEND:
				{
					*pa = make_action_suspend_process();
					if (!*pa) {
						rc = 0;
					}
				}
				break;
			case AT_PROCESS_KILL:
				{
					*pa = make_action_kill_process();
					if (!*pa) {
						rc = 0;
					}
				}
				break;
			case AT_SYSCALL_FAIL:
				{
					int error_code;
					rc = decode_int(buf,
						 	buflen,
							pos,
							&error_code);
					if (rc > 0) {
						*pa = make_action_fail_syscall(
								    error_code);
						if (!*pa) {
							rc = 0;
						}
					}
				}
				break;
			case AT_NIL:
				/* a NULL action_node. */
				*pa = NULL;
				rc = 1;
				break;
			default:
				/* shouldn't happen. */
				rc = 0;
				break;
		}
	}

	return rc;
}


/*
 * Functions for actions manipulation.
 */

/* allocate an action object for logging purposes. */
action_t*
make_action_log(int priority)
{
/* OPTIMIZATION: keep static actions and return pointers to them */

        struct tracker_action* action =
		mem_alloc(sizeof(struct tracker_action));
	if (action) {
		action->type = AT_LOG;
		action->u.priority = priority;
	}

	return action;
}

/* allocate an action object for rule enabling purposes. */
action_t*
make_action_enable_rule(int rule_id)
{
	struct tracker_action* action =
		mem_alloc(sizeof(struct tracker_action));
	if (action) {
		action->type = AT_RULE_ENABLE;
		action->u.rule_id = rule_id;
	}

	return action;
}

/* allocate an action object for rule disabling purposes. */
action_t*
make_action_disable_rule(int rule_id)
{
	struct tracker_action* action =
		mem_alloc(sizeof(struct tracker_action));
	if (action) {
		action->type = AT_RULE_DISABLE;
		action->u.rule_id = rule_id;
	}

	return action;
}

/* allocate an action object for process suspending. */
action_t* make_action_suspend_process()
{
	struct tracker_action* action =
		mem_alloc(sizeof(struct tracker_action));
	if (action) {
		action->type = AT_PROCESS_SUSPEND;
	}

	return action;
}

/* allocate an action object for process killing. */
action_t* make_action_kill_process()
{
	struct tracker_action* action =
		mem_alloc(sizeof(struct tracker_action));
	if (action) {
		action->type = AT_PROCESS_KILL;
	}

	return action;
}

/* allocate an action object for system call failing. */
action_t* make_action_fail_syscall(int error_code)
{
	struct tracker_action* action =
		mem_alloc(sizeof(struct tracker_action));
	if (action) {
		action->type = AT_SYSCALL_FAIL;
		action->u.error_code = error_code;
	}

	return action;
}


/*
 * Action serialization/de-serialization functions.
 */

/* serialize an action node into a buffer, for export     */
/* purposes. returns the size of the serialized data, 0   */
/* if there was an error, -1 if buffer is too small.      */
int
serialize_action_node(action_t* a, unsigned char* buf, int buflen)
{
	unsigned char* pos = buf;
	int rc = do_serialize_action_node(a, buf, buflen, &pos);

	return (rc > 0 ? pos - buf : rc);
}

/* de-serialize an action node from a buffer, for import  */
/* purposes. returns the new allocated action node or     */
/* NULL if an error occuerd.                              */
action_t*
deserialize_action_node(unsigned char* buf, int buflen)
{
	unsigned char* pos = buf;
	action_t* a = NULL;
	int rc = 0;

	if (buflen > 0) {
		rc = do_deserialize_action_node(&a, buf, buflen, &pos);
	}

	return (rc > 0 ? a : NULL);
}

/* print an action node to a "standard" device */
void print_action_node(action_t* a)
{
	char* action_name;

	switch (a->type) {
		case AT_LOG:
			action_name = "log";
			break;
		case AT_RULE_ENABLE:
			action_name = "enable rule";
			break;
		case AT_RULE_DISABLE:
			action_name = "disable rule";
			break;
		case AT_PROCESS_SUSPEND:
			action_name = "suspend process";
			break;
		case AT_PROCESS_KILL:
			action_name = "kill process";
			break;
		case AT_SYSCALL_FAIL:
			action_name = "fail syscall";
			break;
		default:
			action_name = "(unknown)";
			break;
	}
	PPRINT(" Action: %s\n", action_name);
}


/*
 * Functions for tracking rules manipulation.
 */

/* serialize a tracking rule into a buffer, for export   */
/* purposes. returns the size of the serialized data, 0  */
/* if there was an error, -1 if the buffer is too small. */
int serialize_tracking_rule(int rule_id, int syscall_id,
			    action_t* a, filter_node_t* f,
		            unsigned char* buf, int buflen)
{
	unsigned char* pos_1 = buf;
	int rc = encode_int(buf, buflen, &pos_1, rule_id);
	if (rc > 0) {
	    	rc = encode_int(buf, buflen, &pos_1, syscall_id);
		if (rc > 0) {
			int action_size =
		    	serialize_action_node(a, pos_1, buflen-(pos_1-buf));
			if (action_size > 0) {
				unsigned char* pos_2 = pos_1+action_size;
				int filter_size =
					serialize_filter_node(f,
						      	pos_2,
						      	buflen-(pos_2-buf));
				if (filter_size > 0) {
					rc = (pos_2 - buf) + filter_size;
				}
				else {
					rc = filter_size;
				}
			}
			else {
				rc = action_size;
			}
		}
	}

	return rc;
}

/* de-serialize a tracking rule from a buffer, for import */
/* purposes. returns the 1 for success, -1 for failure.   */
/* the given action rule and filter node are filled with  */
/* elements geenrated from the buffer, and should be      */
/* freed by the caller.                                   */
int deserialize_tracking_rule(unsigned char* buf, int buflen, 
			      int* p_rule_id, int* p_syscall_id,
			      action_t** pa, filter_node_t** pf)
{
	unsigned char* pos_1 = buf;
	int rc = decode_int(buf, buflen, &pos_1, p_rule_id);

	if (rc > 0) {
		rc = decode_int(buf, buflen, &pos_1, p_syscall_id);
		if (rc > 0) {
			unsigned char* pos_2 = pos_1;
			rc = do_deserialize_action_node(pa,
							pos_1,
							buflen-(pos_1-buf),
							&pos_2);
			if (rc > 0) {
				unsigned char* pos_3 = pos_2;
				rc = do_deserialize_filter_node(pf,
								pos_2,
								buflen-(pos_2-buf),
								&pos_3);
			}
		}
	}

	return rc;
}

/* create an empty tracking rules list. */
rule_list_t*
create_tracking_rule_list()
{
	rule_list_t* rules = mem_alloc(sizeof(struct rule_list));
	if (rules) {
		rules->num_rules = 0;
	}

	return rules;
}

/* delete the given tracking rule list, freeing any rule found in it. */
void
delete_tracking_rule_list(rule_list_t* rules)
{
	del_all_tracking_rules_in_list(rules);
	mem_free(rules);
}

/* add a new tracking rule. returns the new rule's id */
/* returns '0' if an error occured, '-1' if there is  */
/* already a rule with the given rule ID.             */
int
add_tracking_rule(rule_list_t* rules,
		  int rule_id,
		  filter_node_t* filter,
		  action_t* action)
{
	/* assume failure. */
	int res = 0;

	if (filter && action) {
		/* make sure we don't already have a rule with the same ID. */ 
		int rule_loc = find_tracking_rule(rules, rule_id);
		if (rule_loc >= 0) {
			res = -1;
		}
		else {
			/* add the rule, sorted by rule ID. */
			res = do_add_tracking_rule(rules,
						   rule_id,
						   filter,
						   action);
		}
	}

	return res;
}

/* delete a tracking rule. */
void
del_tracking_rule(rule_list_t* rules, int rule_id)
{
    int i = find_tracking_rule(rules, rule_id);

    if (i >= 0 && i < rules->num_rules) {
	struct tracker_rule* rule = rules->rules[i];
	int j;

	/* we're removing a rule - close the 'gap' created in the list. */
	for (j = i; j < rules->num_rules-1; j++) {
		rules->rules[j] = rules->rules[j+1];
	}
	if (rule->filter) {
	    free_filter_node(rule->filter);
	}
	mem_free(rule);
	rules->num_rules--;
    }
}

/* deletes all tracking rules in the list. */
void
del_all_tracking_rules_in_list(rule_list_t* rules)
{
    int i;

    for (i = 0; i < rules->num_rules; i++) {
	if (rules->rules[i]) {
	    if (rules->rules[i]->filter) {
	        free_filter_node(rules->rules[i]->filter);
	    }
	    mem_free(rules->rules[i]);
	    rules->rules[i] = NULL;
	}
    }
    rules->num_rules = 0;
}

/* check if there is a rule with the given ID. */
/* returns '1' if there is, '0' otehrwise.     */
int
has_tracking_rule(rule_list_t* rules, int rule_id)
{
    int i = find_tracking_rule(rules, rule_id);

    return (i >= 0 ? 1 : 0);
}

/* print out the list of current tracking rules. */
void print_tracking_rules(rule_list_t* rules)
{
	int i;

	PPRINT("\nActive Tracking Rules List (%d):\n", rules->num_rules);
	for (i=0; i < rules->num_rules; i++) {
	     /* skip empty rules */
	     if (!rules->rules[i])
		  continue;
	     PPRINT(" Rule id: %d\n", rules->rules[i]->rule_id);
	     print_action_node(&rules->rules[i]->action);
	     PPRINT(" Filter:\n");
	     print_filter_node(rules->rules[i]->filter);
	}
	PPRINT("\n");
}

/*
 * Functions used by tracking functions to perform filtering.
 */

/* filter an incoming system call.                                          */
/* accepts a system call ID and its (var_args) parameters.                  */
/* returns a tracker_action structure if there was a match, NULL otherwise. */
action_t* filter_syscall(rule_list_t* rules, int syscall_id, ...)
{
	int i;
	action_t* action = NULL;
	filter_node_t* result = make_bool_val(TF_FALSE);

	for (i=0; i<rules->num_rules; i++) {
		if (rules->rules[i] && rules->rules[i]->filter) {
			exp_error_t rc =
			    eval_filter_node(rules->rules[i]->filter, result);
			if (rc == ERR_NONE && result->data_type == DT_BOOL &&
			    result->u.constant.u.b)
				break;
		}
	}

	if (i < rules->num_rules) {
		action = &rules->rules[i]->action;
	}

	return action;
}

/* log the syscall that matched one of the rules in 'rules' */
static void 
tracker_do_logging(rule_list_t* rules, 
		   int syscall_id, 
		   const char* param_spec, 
		   va_list ap)
{
    PPRINT(KERN_NOTICE "syscall '%d' invoked %s", syscall_id,
	   param_spec && *param_spec ? "with params: ": "");
    
    if (param_spec) {
	while (*param_spec) {
	    switch (*param_spec++){
	    case 'c': PPRINT("%c", (char)va_arg(ap, int)); break;
            case 'h': PPRINT("%hi", (short)va_arg(ap, int)); break; 
	    case 'i': PPRINT("%d", va_arg(ap, int)); break;
	    case 'l': PPRINT("%ld", va_arg(ap, long)); break;
	    case 's': PPRINT("%s", va_arg(ap, char*)); break;
	}
	    if (*param_spec)
		PPRINT(", ");
	}
    }
    PPRINT("\n");
}

/* perform an action in the current system call's context. */
/* accepts a system call ID and its (var_args) parameters. */
void
tracker_perform_action(rule_list_t* rules,
		       action_t* action,
		       int syscall_id,
		       const char* param_spec,
		       ...)
{
    if (action) {
	/* TODO TODO - handle other action types. */
	/* also handle logging with varargs.      */
	switch (action->type) {
	case AT_LOG:
	{
	    va_list ap;
	    va_start (ap, param_spec);
	    tracker_do_logging(rules, syscall_id, param_spec, ap);
	    va_end(ap);
	    break;
	}		
	default:
	    PPRINT(KERN_NOTICE "syscall '%d' invoked - unknown action %d\n",
		   syscall_id, action->type);
	    break;
	}
    }
}

