#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>

#include "tester.h"

#include "sct_ctrl_lib/sct_ctrl_lib.h"

static unsigned int rule_id = 0;
extern sct_ctrl_t ctrl;
extern char* progname;

static void
register_module_passwd_rule(unsigned int rid)
{
    filter_node_t* f0 = make_var_indexed_val(VT_PARAM, 0, DT_STRING);
    filter_node_t* f1 = make_string_val("passwd");
    filter_node_t* f2 = make_pattern_match(f0, f1);
    action_t* a1 = make_action_log(4);
    
    if (!sct_ctrl_add_rule(ctrl, rid, 10, f2, a1, 1))
	die ("failed to upload rule '%d'\n", rid);
}

int
test_simple_syscall_usage(void)
{
    unsigned int num_tests = 100;
    int rc = fork();
    if (rc == -1)
	die("'%s': fork failed\n", __FUNCTION__);

    /* both father and child execute this */ 
    while (--num_tests){ 
	int delay = 1+(int) (10.0 * rand() /(RAND_MAX+1.0));
	creat("passwd", S_IRUSR | S_IWUSR);
	usleep(delay);
	remove("passwd");
    }
    
    if (rc == 0) /* child */
	_exit(0);

    return 1;
}

int 
test_rule_register_deregister(void)
{
    unsigned int num_tests = 100;
    int rc;

    if (!sct_ctrl_del_all_rules(ctrl))
	die("'%s': failed to delete all module rules\n");

    rc = fork();
    if (rc == -1)
	die("'%s': fork failed\n", __FUNCTION__);

    if (rc == 0){ /* child code */
	++rule_id;
    }
    
    /* both father and child execute this */ 
    while (--num_tests){ 
	int delay = 1+(int) (5.0 * rand() /(RAND_MAX+1.0));
	unsigned int rid = (rule_id += 2);
	register_module_passwd_rule(rid);
	usleep(delay);
	if (!sct_ctrl_del_rule(ctrl, rid, 10))
	    die ("'%s': failed to delete rule\n", __FUNCTION__);
    }
    
    if (rc == 0) /* child */
	_exit(0);

    sct_ctrl_print_all_rules(ctrl);

    return 1;
    
}

static void 
all_syscalls_register_rules(void)
{
    filter_node_t* p0 = make_var_indexed_val(VT_PARAM, 0, DT_STRING);
    filter_node_t* p1 = make_var_indexed_val(VT_PARAM, 1, DT_STRING);

    filter_node_t* prog = make_string_val(progname);
    filter_node_t* progmatch0 = make_pattern_match(p0, prog);
/*     filter_node_t* progmatch1 = make_pattern_match(p1, prog); */

    filter_node_t* tmp = make_string_val("tmp");
    filter_node_t* tmpmatch0 = make_pattern_match(p0, tmp);
    filter_node_t* tmpmatch1 = make_pattern_match(p1, tmp);

/*    filter_node_t* dir = make_string_val("."); */
/*    filter_node_t* dirmatch0 = make_pattern_match(p0, dir); */

    action_t* a1 = make_action_log(4);

    if (!sct_ctrl_add_rule(ctrl, ++rule_id, 5, progmatch0, a1, 1))
	die ("failed to upload rule '%d'\n", rule_id);

    if (!sct_ctrl_add_rule(ctrl, ++rule_id, 9, tmpmatch1, a1, 1))
	die ("failed to upload rule '%d'\n", rule_id);

    if (!sct_ctrl_add_rule(ctrl, ++rule_id, 10, tmpmatch0, a1, 1))
	die("failed to upload rule '%d'\n", rule_id);

    if (!sct_ctrl_add_rule(ctrl, ++rule_id, 12, progmatch0, a1, 1))
	die("failed to upload rule '%d'\n", rule_id);

    if (!sct_ctrl_add_rule(ctrl, ++rule_id, 15, tmpmatch0, a1, 1))
	die("failed to upload rule '%d'\n", rule_id);
    
    if (!sct_ctrl_add_rule(ctrl, ++rule_id, 33, progmatch0, a1, 1))
	die("failed to upload rule '%d'\n", rule_id);

    /* TODO: add 'kill' rule */

    if (!sct_ctrl_add_rule(ctrl, ++rule_id, 39, tmpmatch0, a1, 1))
	die("failed to upload rule '%d'\n", rule_id);
    
    if (!sct_ctrl_add_rule(ctrl, ++rule_id, 40, tmpmatch0, a1, 1))
	die("failed to upload rule '%d'\n", rule_id);

    sct_ctrl_print_all_rules(ctrl);
    sleep(3);
}

static void 
all_syscalls_deregister_rules(void)
{
    if (!sct_ctrl_del_all_rules(ctrl))
	die("failed to delete all module rules");
}

int
test_all_syscalls_usage(void)
{
    const unsigned int num_syscalls = 9;
    unsigned int times_to_run = 600;
    unsigned int i;
    int rc;

    all_syscalls_register_rules();
    
    for (i = 0; i < num_syscalls; ++i){
	/* link/unlink and mkdir/rmdir are handled by one process
	   each pair, so avoid creating a new process */
	if ((i == 2) || (i== 8)) {
	    ++i;
	    continue;
        }
	rc = fork();
	if (rc == -1)
	    die("'%s': fork failed\n", __FUNCTION__);
	if (rc == 0) 
	    break;
    }
    /* parent code */
    if (rc != 0){
	unsigned int children = num_syscalls;
	while (children--)
	    wait(NULL);
    } else {
        /* children code */
        while (--times_to_run){
	    int delay = 1+(int) (10.0 * rand() /(RAND_MAX+1.0));
	    usleep(delay);

	    switch (i){
	    case 0: 	/* open */
	    {
	        int fd;
	        if ((fd = open(progname, O_RDONLY)) == -1){
		    perror("open");
		    die ("'%s': open failed\n", __FUNCTION__);
	        }
	        close(fd);
	        if (times_to_run % 10 == 0){
	            fprintf(stderr, ".");
		    fflush(stderr);
	        }
	        break;
	    }
	    case 1: /* link */
	    case 2: /* unlink */
	    {
	        if (link(progname, "tmplink") == -1){
		    perror("link");
		    die ("'%s': link failed\n", __FUNCTION__);
	        }
	        unlink("tmplink");
	        break;
	    } 
	    case 3: /* chdir */
	    {
	        static char cwd[1024];
	        getcwd(cwd, sizeof(cwd));
	        if (chdir("..") == -1){
		    perror("chdir");
		    die ("'%s': chdir \"..\" failed\n", __FUNCTION__);
	        }
	        if (chdir(cwd) == -1){
		    perror("chdir");
		    die ("'%s': chdir \"%s\" failed\n", __FUNCTION__, cwd);
	        }
	        break;
	    }
	    case 4: /* chmod */
	    {
	        if (chmod(progname, S_IXUSR) == -1){
		    perror("chmod");
		    die ("'%s': chmod failed\n", __FUNCTION__);
	        }
	        break;
	    }
	    case 5: /* access */
	    {
	        if (access(progname, X_OK) == -1){
		    perror("access");
		    die ("'%s': access failed\n", __FUNCTION__);
	        }
	        break;
	    }
	    case 6: /* kill */
	    {   
	        if (kill(getpid(), 0) == -1){
		    perror("kill");
		    die ("'%s': kill failed\n", __FUNCTION__);
	        }
	        break;
	    }
	    case 7: /* mkdir */
	    case 8: /* rmdir */
	    {
	        if (mkdir("tmpdir", S_IRWXU) == -1){
		    perror("mkdir");
		    die ("'%s': mkdir failed\n", __FUNCTION__);
	        }
	        if (rmdir("tmpdir") == -1){
		    perror("rmdir");
		    die ("'%s': rmdir failed\n", __FUNCTION__);
	        }
	        break;
	    }
	    }
        }
        /* child exits here. */
	fprintf(stderr, "child '%d' finished\n", i);
	_exit(0);
    }
    
    all_syscalls_deregister_rules();

    return 1;
}

